/*
 * Copyright (C) 2013  Alexandre Magaz Graça
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.uoc.sig.indoorlocator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import com.wifislam.indoorlocation.IndoorLocation;
import com.wifislam.indoorlocation.IndoorLocationListener;
import com.wifislam.indoorlocation.IndoorLocationManager;
import com.wifislam.indoorlocation.IndoorLocationManager.IndoorLocationStatus;
import com.wifislam.indoorlocation.IndoorLocationManager.LocationChoice;
import com.wifislam.indoorlocation.IndoorLocationManager.LocationChoice.LocationType;
import com.wifislam.indoorlocation.IndoorLocationView;

import android.net.wifi.WifiManager;
import android.os.Bundle;
import android.os.Environment;
import android.os.IBinder;
import android.app.Activity;
import android.content.ComponentName;
import android.content.Context;
import android.content.ServiceConnection;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.widget.TextView;
import android.widget.Toast;

/**
 * Main and only activity to show the position on the floor plan.
 * 
 * @author Àlex Magaz Graça
 */
public class MapActivity extends Activity implements ServiceConnection, IndoorLocationListener{

	// File where positioning information is stored on user request
	private static final String MEASURES_FILE = "measures.csv";
	/* ID returned by IndoorLocation.getLocationId() when no location has been
	  determined yet. */
	private static final String NO_LOCATION_ID = "OffMap";
	private static final String TAG = "IndoorLocator"; // Log tag

	
	private IndoorLocationManager locationManager;
	private String currentLocationId;
	private IndoorLocation lastLocation;
	
	
    @Override
    public void onCreate(Bundle savedInstanceState) {
    	IndoorLocationManager.Configuration configuration;
    	View locationView;
    	LocationChoice locationChoice; 
    	
        super.onCreate(savedInstanceState);
        
        currentLocationId = NO_LOCATION_ID;
        configuration = new IndoorLocationManager.Configuration();
        configuration.enableAutoDownload = true;
        configuration.enableFloorplanDownloads = true;
        locationChoice = new LocationChoice(LocationType.LOCATION_GROUP_ID, "1800");
        
        locationManager = new IndoorLocationManager(this, configuration,
        												this, locationChoice);
        
        setContentView( R.layout.activity_map );
        locationView = locationManager.createLocationView(this).getView();
        ViewGroup mapContainer = (ViewGroup) findViewById( R.id.map_container );
        mapContainer.addView(locationView);
        setContentView(mapContainer);
    }

    @Override
	public void onServiceConnected(ComponentName name, IBinder service) {
    	//TODO: substitute <API key> bellow with your WiFiSLAM API key
    	locationManager.setApiKey("<API key>");
    	locationManager.requestLocationUpdates(this);
    	locationManager.startLocalizer();
	}

	@Override
	public void onServiceDisconnected(ComponentName name) {
		Toast.makeText(this,
						"Connection lost with location service.", 
						Toast.LENGTH_LONG).show();
	}

	@Override
    public boolean onCreateOptionsMenu(Menu menu) {
        getMenuInflater().inflate(R.menu.activity_map, menu);
        return true;
    }

	@Override
	protected void onDestroy() {
		super.onDestroy();
		locationManager.unbind();
	}

	@Override
	public void onIndoorLocationChanged(IndoorLocation location) {
		// If the location has changes, notify the user about the new location.
		if ( !location.getLocationId().equals(currentLocationId) ) {
			currentLocationId = location.getLocationId();
			Toast.makeText(this,
							location.getLocationDescription(), 
							Toast.LENGTH_LONG).show();
		}                                                                                                                                               

		// To be able to save measures when the user asks with "Take Measure"
		lastLocation = location;
		
		double relativeX = location.getRelativeX();
		double relativeY = location.getRelativeY();
		double latitude = location.getLatitude();
		double longitude = location.getLongitude();
		TextView positionView = (TextView) findViewById(R.id.status);
		
		// Don't display anything until we get an actual position
		if ( !Double.isNaN(relativeX) ) {		
			// Calculate distance to the upper left corner of the map (0, 0).
			double distance = Math.sqrt( Math.pow(relativeX, 2.0)
										+ Math.pow(relativeY, 2.0) );
			String distanceText = String.format("distance=%.2fm", distance);
			String geoPositionText = String.format("lat=%.5f lon=%.5f", 
													latitude,
													longitude);
			String accuracyText = String.format("accuracy=%.2fm", 
												location.getAccuracy());
			
			positionView.setText(distanceText + "; " 
								+ geoPositionText + "; " 
								+ accuracyText);
		} else
			positionView.setText("");
	}

	@Override
	public void onStatusChanged(IndoorLocationStatus status) {
		if ( status == IndoorLocationStatus.LOGGING_IN ) {
			TextView positionView = (TextView) findViewById(R.id.status);
			positionView.setText("Logging in...");
		}
		// The other statuses are not useful by the moment
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch ( item.getItemId() ) {
			case R.id.menu_take_measure:
			    storeMeasure();
			    return true;
			case R.id.menu_disassociate_wifi:
				disassociateWifi();
			    return true;
			default:
				return super.onOptionsItemSelected(item);
		}
	}
	
	/**
	 * Stores a new record with current positioning information into a CSV file.
	 * 
	 * The file should be stored in the device's external storage and 
	 * in the private directory of the application. This should be:
	 * 
	 * <code>/Android/data/edu.uoc.sig.indoorlocator/files/measures.csv</code>
	 * 
	 * If the file doesn't exists, it is created. From this point, 
	 * records are always appended to the end of the file.
	 * 
	 * The format of the file can be seen in the following example:
	 * 
	 * <code>
	 * 2012-12-10 18:53:23,1st floor,26.65,39.68,423,776,630,938,41.608545,0.624483,2.58
	 * 2012-12-10 18:55:30,1st floor,28.77,39.84,464,779,680,942,41.608531,0.624466,2.52
	 * 2012-12-11 19:31:54,1st floor,6.51,48.42,29,947,154,1145,41.608737,0.624548,2.73
	 * </code>
	 * 
	 * Where the values written in each record correspond, from 
	 * left to right, to the following information:
	 * 
	 * <ul>
	 * 	<li>timestamp</li>
	 *  <li>location</li>
	 *  <li>horizontal meters coordinates</li>
	 *  <li>vertical meters coordinates</li>
	 *  <li>horizontal pixels coordinates</li>
	 *  <li>vertical pixels coordinates</li>
	 *  <li>horizontal pixels coordinates (from scaled floor plan)</li>
	 *  <li>vertical pixels coordinates (from scaled floor plan)</li>
	 *  <li>latitude</li>
	 *  <li>longitude</li>
	 *  <li>accuracy</li>
	 * </ul>
	 * 
	 * Note that all information is always written in US locale 
	 * independently of the actual user locale. This avoids, for 
	 * example, having to quote decimal values (Spanish locale 
	 * would use ',' as decimal separator).
	 */
	private void storeMeasure() {
		String state = Environment.getExternalStorageState();

        if (Environment.MEDIA_MOUNTED.equals(state)) {
        	DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
			String measureCsv = String.format(
								Locale.US,
								"%s,%s,%.2f,%.2f,%d,%d,%d,%d,%.6f,%.6f,%.2f\n", 
								dateFormat.format(new Date()),
								lastLocation.getLocationDescription(),
								lastLocation.getRelativeX(),
								Math.abs(lastLocation.getRelativeY()),
								(int) lastLocation.getFloorplanX(),
								(int) lastLocation.getFloorplanY(),
								(int) lastLocation.getScaledFloorplanX(),
								(int) lastLocation.getScaledFloorplanY(),
								lastLocation.getLatitude(),
								lastLocation.getLongitude(),
								lastLocation.getAccuracy());
            File file = new File(getExternalFilesDir(null), MEASURES_FILE);
            
            try {
				OutputStream os = new FileOutputStream(file, true);
				os.write(measureCsv.getBytes());
				os.close();
			} catch (FileNotFoundException e) {
				Toast.makeText(this,
						"Could not store the measure: Error opening the measures file.", 
						Toast.LENGTH_LONG).show();
			} catch (IOException e) {
				Toast.makeText(this,
						"Error storing the measure: " + e.getLocalizedMessage(), 
						Toast.LENGTH_LONG).show();
			}
        } else {
			Toast.makeText(this,
					"Could not store the measure: external storage not available.", 
					Toast.LENGTH_LONG).show();
        }		
	}
	
	/**
	 *  Disassociates from the current WiFi AP.
	 *  
	 *  It usually improves scan frequency. 
	 */
	private void disassociateWifi() {
		WifiManager wifiManager = (WifiManager) 
									getSystemService(Context.WIFI_SERVICE);
		wifiManager.disableNetwork(
							wifiManager.getConnectionInfo().getNetworkId() );
	}
}
